import unittest
import numpy as np
import torch
import torch.nn.functional as F
from smap import *

# the test case
class SMapUTestCase(unittest.TestCase):
    def setUp(self):
        self.device = "cpu"
        self.n = 2
        self.img_shape = [128, 256]
        self.camera = np.array([[2304.5479, 0,  1686.2379], 
                                [0, 2305.8757, -0.0151],
                                [0, 0, 1.]], dtype=np.float32)
        self.panel = list(np.where(np.ones([self.img_shape[0], self.img_shape[1]])))
        self.panel[0] = self.panel[0] + .5
        self.panel[1] = self.panel[1] + .5
        self.smap = SMap(self.n, self.img_shape[0], self.img_shape[1], self.camera, self.device).to(self.device)
        self.smap3x3 = self.smap.smap3x3

    def test_SMap_forward(self):
        zoom = 0
        activated_coords = [np.array([np.random.choice(range(self.img_shape[0]), size=None)]), np.array([np.random.choice(range(self.img_shape[1]), size=None)])]
        traverse_diff = np.array([0., -0., 0.]).reshape(1,3)
        mask = 0.*(np.load(f"./tests/vtest_data/smap/mask.npy")[0,0,:,:])
        for i, (r, c) in enumerate(zip(*activated_coords)):
            mask[r, c] = 1.
        input_mask = torch.from_numpy(mask).float().to(self.device).reshape(1,1, self.img_shape[0], self.img_shape[1])
        z = 1.
        input_repr = (z*input_mask)
        input_repr = utils.to_3d(input_repr.reshape(1,1, self.img_shape[0], self.img_shape[1]), self.img_shape[0], self.img_shape[1], self.panel, self.img_shape, self.img_shape, self.smap3x3.camera_matrix_inv, self.device).reshape(1,3, self.img_shape[0], self.img_shape[1])
        
        actual = self.smap(torch.cat([input_repr, input_mask], dim=1), None, zoom)
        actual = (actual[:,-1:,(actual.shape[-2]-self.img_shape[0])//2:-(actual.shape[-2]-self.img_shape[0])//2, (actual.shape[-1]-self.img_shape[1])//2:-(actual.shape[-1]-self.img_shape[1])//2])
        actual = actual.detach().cpu().numpy().reshape(self.img_shape[0], self.img_shape[1])
        
        expected_activated_coords = np.where(mask>0.)
        for i, (r, c) in enumerate(zip(*activated_coords)):
            coord = (input_repr[0,:,r, c]).numpy().reshape(1,3)
            new_coord = coord+traverse_diff
            new_coord = torch.einsum('xz,yz->xy', torch.from_numpy(new_coord).float(), torch.from_numpy(self.camera).float()).cpu().numpy()
            
            new_r = new_coord[0,1]/new_coord[0,-1]-.5
            new_c = new_coord[0,0]/new_coord[0,-1]-.5
            expected_activated_coords[0][i] = round(new_r)
            expected_activated_coords[1][i] = round(new_c)
        expected = np.zeros_like(actual)
        expected[expected_activated_coords] = 1.
        
        try:
            np.testing.assert_almost_equal(actual, expected, decimal=3, err_msg='SMap forward test failed.')
        except Exception as e:
            raise e

if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn
import numpy as np

class NoneBot(nn.Module):
    def __init__(self, module=None):
        super(NoneBot, self).__init__()
        self.module = module
        
    def forward(self, x):
        if self.module is not None:
            return self.module(x)
        return x+0
    
class TestBot_Target(nn.Module):
    def __init__(self, module=None, name="out"):
        super(TestBot_Target, self).__init__()
        def get_activation_grad(name='out'):
            def hook(module, grad_inputs, grad_outputs):
                pass
            return hook
        self.testcase = None
        self.module = NoneBot()
        if module is not None:
            self.module = module
        self.target_representation = None
        self.name = name
        self.module.register_backward_hook(get_activation_grad(self.name))
        
    def forward(self, x):
        h, w = x.shape[-2], x.shape[-1]
        self.target_representation = x.detach().cpu().numpy().reshape(h, w)
        if self.testcase is None:
            np.save(self.testcase.out_path+"target_representation.npy", self.target_representation)
        else:
            np.save(self.testcase.out_path+self.testcase.name+"_target_representation.npy", self.target_representation)
        return self.module(x)
    
class TestBot_In(nn.Module):
    def __init__(self, module=None, offset_in=0, name="in", connet2name="out"):
        super(TestBot_In, self).__init__()
        def get_activation_grad(name, connet2name="out"):
            def hook(module, grad_inputs, grad_outputs):
                if name is not None:
                    # Lấy grad_out: shape [N, C_out, H_out, W_out]
                    grad_out, grad_in = None, None
                    for grad in grad_inputs:
                        if grad is not None:
                            grad_in = grad
                    H_in, W_in = (grad_in.shape[-2]), (grad_in.shape[-1])
                    grad_in = (grad_in[:,:,offset_in,:,:]).reshape(H_in, W_in)
                    
                    self.testcase.activation_gradients[name] = grad_in.cpu().numpy()
                    
                    self.testcase.gradient_flows[(name, connet2name)] = None
                    import pickle
                    flow_info = {"activation_gradients": self.testcase.activation_gradients, 
                                 "gradient_flows": self.testcase.gradient_flows}
                    if self.testcase is None:
                        np.save(self.testcase.out_path+"input_representation.npy", self.input_representation.reshape(H_in, W_in))
                        with open(self.testcase.out_path+'flow_info.pkl', 'wb') as f:
                            pickle.dump(flow_info, f)
                    else:
                        np.save(self.testcase.out_path+self.testcase.name+"_input_representation.npy", self.input_representation.reshape(H_in, W_in))
                        with open(self.testcase.out_path+self.testcase.name+"_flow_info.pkl", 'wb') as f:
                            pickle.dump(flow_info, f)
            return hook
        self.testcase = None
        self.module = NoneBot()
        if module is not None:
            self.module = module
        self.name = name
        self.connet2name = connet2name
        self.input_representation = None
        self.module.register_backward_hook(get_activation_grad(self.name, self.connet2name))
        
    def forward(self, x, mask):
        h, w = mask.shape[-2], mask.shape[-1]
        self.input_representation = mask.detach().cpu().numpy().reshape(h, w)
        return self.module(x)
    
class TestBot_Out(nn.Module):
    def __init__(self, module=None, offset_out=4, name="in", connet2name="out"):
        super(TestBot_Out, self).__init__()
        def get_activation_grad(name, connet2name="out"):
            def hook(module, grad_inputs, grad_outputs):
                if name is not None:
                    # Lấy grad_out: shape [N, C_out, H_out, W_out]
                    grad_out, grad_in = None, None
                    for grad in grad_outputs:
                        if grad is not None:
                            grad_out = grad
                    H_out, W_out = (grad_out.shape[-2]), (grad_out.shape[-1])
                    
                    grad_out = (grad_out[:,offset_out,:,:]).reshape(H_out, W_out)
                    
                    self.testcase.activation_gradients[connet2name] = grad_out.cpu().numpy()
                    
                    self.testcase.gradient_flows[(name, connet2name)] = None
            return hook
        self.testcase = None
        self.module = NoneBot()
        if module is not None:
            self.module = module
        self.name = name
        self.connet2name = connet2name
        self.module.register_backward_hook(get_activation_grad(self.name, self.connet2name))
        
    def forward(self, x):
        return self.module(x)

class TestCase():
    def __init__(self, name="", testbot_in=None, testbot_out=None, testbot_target=None):
        self.out_path = "./tests/vtest_data/output/"
        self.name = name
        self.activation_gradients = {}
        self.gradient_flows = {}
        self.testbot_in = testbot_in
        self.testbot_out = testbot_out
        self.testbot_target = testbot_target
        self.testbot_in.testcase = self
        self.testbot_out.testcase = self
        self.testbot_target.testcase = self
        
    def get_testbot_in(self):
        return self.testbot_in
    
    def get_testbot_out(self):
        return self.testbot_out
    
    def get_testbot_target(self):
        return self.testbot_target



import unittest
import numpy as np
import torch
import torch.nn.functional as F
from smap import *

# the test case
class SMap3x3UTestCase(unittest.TestCase):
    def setUp(self):
        self.device = "cpu"
        self.img_shape = [8, 12]
        self.camera = np.array([[2304.5479, 0,  1686.2379], 
                                [0, 2305.8757, -0.0151],
                                [0, 0, 1.]], dtype=np.float32)
        self.panel = list(np.where(np.ones([self.img_shape[0], self.img_shape[1]])))
        self.smap3x3 = SMap3x3(self.img_shape[0], self.img_shape[1], self.camera, self.device).to(self.device)

    def test_to_3d3x3(self):
        depth_map = torch.from_numpy(np.zeros((self.img_shape[0], self.img_shape[1]))).float().to(self.device)
        
        active_point_img_coords = [np.random.choice(range(self.img_shape[0]),size=None), np.random.choice(range(self.img_shape[1]),size=None)]
        depth_map[active_point_img_coords[0],active_point_img_coords[1]] = 1e3*(np.random.rand(1)[0])
        offsetx, offsety = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
        
        actual = (utils.to_3d3x3(depth_map.reshape(1,1,self.img_shape[0],self.img_shape[1]), self.img_shape[0], self.img_shape[1], self.panel, self.img_shape, self.img_shape, self.smap3x3.camera_matrix_inv, self.device).reshape(3,3,self.img_shape[0],self.img_shape[1],3)[offsetx,offsety,active_point_img_coords[0],active_point_img_coords[1],:]).cpu().numpy()
        pointy = (active_point_img_coords[1])+(offsety-1)
        if pointy<0 or pointy>=(self.img_shape[1]):
            pointy=0
        pointx = (active_point_img_coords[0])+(offsetx-1)
        if pointx<0 or pointx>=(self.img_shape[0]):
            pointx=0
        expected = torch.einsum("x,yx->y", (depth_map[active_point_img_coords[0], active_point_img_coords[1]])*torch.from_numpy(np.array([pointy,pointx,1.])).float(), torch.from_numpy(np.linalg.inv(self.camera))).numpy()
        try:
            np.testing.assert_allclose(actual, expected, atol=1e-3,
                                       err_msg="Transforming from depth to 3D coordinates failed.")
        except Exception as e:
            print(f"active_point_img_coords: {active_point_img_coords}")
            print(f"Activated point of depth_map: {depth_map[active_point_img_coords[0],active_point_img_coords[1]]}")
            print(f"offsetx: {offsetx}")
            print(f"offsety: {offsety}")
            raise e
            
    def test_agg_factor_only(self):
        active_point_img_coords0 = [np.random.choice(range(self.img_shape[0]),size=None), np.random.choice(range(self.img_shape[1]),size=None)]
        active_point_img_coords1 = [np.random.choice(range(self.img_shape[0]),size=None), np.random.choice(range(self.img_shape[1]),size=None)]
        offsetx0, offsety0 = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
        offsetx1, offsety1 = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
        first_value = 1e3*(np.random.rand(1)[0])
        second_value = 1e3*(np.random.rand(1)[0])
        unfolded_depth_map = torch.from_numpy(np.zeros((3,3, self.img_shape[0], self.img_shape[1]))).float().to(self.device)
        unfolded_depth_map[offsetx0,offsety0,active_point_img_coords0[0],active_point_img_coords0[1]] = first_value
        unfolded_depth_map[offsetx1,offsety1,active_point_img_coords1[0],active_point_img_coords1[1]] = second_value
        
        factor = 1e15*(np.random.rand(1)[0])
        actual = self.smap3x3.agg(unfolded_depth_map.reshape(1,1,3,3,1, self.img_shape[0], self.img_shape[1])
        , factor=factor).cpu().numpy()
        
        expected = torch.from_numpy(factor*np.ones((3,3, self.img_shape[0], self.img_shape[1]))).float().to(self.device).numpy()
        pointx0 = active_point_img_coords0[0]+offsetx0-1
        pointy0 = active_point_img_coords0[1]+offsety0-1
        pointx1 = active_point_img_coords1[0]+offsetx1-1
        pointy1 = active_point_img_coords1[1]+offsety1-1
        if pointx0<(self.img_shape[0]) and pointy0<(self.img_shape[1]):
            expected[offsetx0,offsety0,pointx0,pointy0] = first_value
        if pointx1<(self.img_shape[0]) and pointy1<(self.img_shape[1]):
            expected[offsetx1,offsety1,pointx1,pointy1] = second_value
        expected = expected.reshape(1,1,3*3,1,1, self.img_shape[0], self.img_shape[1])
        
        try:
            np.testing.assert_allclose(actual, expected,
                                       err_msg="Transforming to absolute-alignment representation failed.")
        except Exception as e:
            print(f"factor: {factor}")
            print(f"active_point_img_coords0: {active_point_img_coords0}")
            print(f"active_point_img_coords1: {active_point_img_coords1}")
            print(f"offsets0: {(offsetx0, offsety0)}")
            print(f"offsets1: {(offsetx1, offsety1)}")
            print(f"first_value: {first_value}")
            print(f"second_value: {second_value}")
            raise e
            
    def test_agg_ind(self):
        active_point_img_coords0 = [np.random.choice(range(self.img_shape[0]),size=None), np.random.choice(range(self.img_shape[1]),size=None)]
        active_point_img_coords1 = [np.random.choice(range(self.img_shape[0]),size=None), np.random.choice(range(self.img_shape[1]),size=None)]
        offsetx0, offsety0 = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
        offsetx1, offsety1 = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
        channel_num = np.random.choice([1,4],size=None)
        first_values = np.random.rand(channel_num)
        first_values[:3] = 1e3*(first_values[:3])
        second_values = np.random.rand(channel_num)
        second_values[:3] = 1e3*(second_values[:3])
        unfolded_depth_map = torch.from_numpy(np.zeros((3,3,channel_num, self.img_shape[0], self.img_shape[1]))).float().to(self.device)
        unfolded_depth_map[offsetx0,offsety0,:channel_num,active_point_img_coords0[0],active_point_img_coords0[1]] = torch.from_numpy(first_values)
        unfolded_depth_map[offsetx1,offsety1,:channel_num,active_point_img_coords1[0],active_point_img_coords1[1]] = torch.from_numpy(second_values)
        
        factor = 1e15*(np.random.rand(1)[0])
        pointx0 = active_point_img_coords0[0]+offsetx0-1
        pointy0 = active_point_img_coords0[1]+offsety0-1
        pointx1 = active_point_img_coords1[0]+offsetx1-1
        pointy1 = active_point_img_coords1[1]+offsety1-1
        
        _ind = self.smap3x3.agg((unfolded_depth_map[:,:,:1,:,:]).reshape(1,1,3*3,1,1, self.img_shape[0], self.img_shape[1]), factor=1e7)
        if channel_num==4:
            _ind = self.smap3x3.agg((unfolded_depth_map[:,:,2:3,:,:]).reshape(1,1,3*3,1,1, self.img_shape[0], self.img_shape[1]), factor=1e7)
        _ind = torch.min(_ind,dim=2,keepdim=True).indices
        ind = 4*torch.ones_like(_ind)
        if pointx0>=0 and pointx0<(self.img_shape[0]) and pointy0>=0 and pointy0<(self.img_shape[1]):
            ind[0,0,0,0,0,pointx0,pointy0] = _ind[0,0,0,0,0,pointx0,pointy0]
        if pointx1>=0 and pointx1<(self.img_shape[0]) and pointy1>=0 and pointy1<(self.img_shape[1]):
            ind[0,0,0,0,0,pointx1,pointy1] = _ind[0,0,0,0,0,pointx1,pointy1]
        ind = F.one_hot(ind, num_classes=3*3).reshape(1,-1,1,1, self.img_shape[0], self.img_shape[1],3*3).permute(0,1,6,2,3,4,5).reshape(1,-1,3*3,1,1, self.img_shape[0], self.img_shape[1])
        ind = (ind>.5)
        actual = self.smap3x3.agg(unfolded_depth_map.reshape(1,1,3,3,channel_num, self.img_shape[0], self.img_shape[1]), ind=ind, factor=factor).cpu().numpy()
        
        expected = torch.from_numpy(factor*np.ones((channel_num, self.img_shape[0], self.img_shape[1]))).float().to(self.device).numpy()
        first_value = (first_values[-1])
        second_value = (second_values[-1])
        if channel_num==4:
            first_value = (first_values[-2])
            second_value = (second_values[-2])
        if pointx1>=0 and pointx1<(self.img_shape[0]) and pointy1>=0 and pointy1<(self.img_shape[1]):
            expected[:,pointx1,pointy1] = torch.from_numpy(second_values)
        if pointx0>=0 and pointx0<(self.img_shape[0]) and pointy0>=0 and pointy0<(self.img_shape[1]):
            if pointx0==pointx1 and pointy0==pointy1:
                expected[:,pointx0,pointy0] = torch.from_numpy(first_values+second_values)
            else:
                expected[:,pointx0,pointy0] = torch.from_numpy(first_values)
        expected = expected.reshape(1,1,1,1,channel_num, self.img_shape[0], self.img_shape[1])
        try:
            np.testing.assert_allclose(actual, expected,
                                       err_msg="Aggregating absolute-alignment representation failed.")
        except Exception as e:
            print(f"factor: {factor}")
            print(f"active_point_img_coords0: {active_point_img_coords0}")
            print(f"active_point_img_coords1: {active_point_img_coords1}")
            print(f"offsets0: {(offsetx0, offsety0)}")
            print(f"offsets1: {(offsetx1, offsety1)}")
            print(f"first_value: {first_value}")
            print(f"second_value: {second_value}")
            raise e
            
    def test_prepare_flows_for_coord(self):
        cases = ["blocked", "random"]
        case_id = np.random.choice(len(cases),size=None)
        
        # With the assumption that the mask at the activated point is 1. 
        height = self.img_shape[0]+2
        width = self.img_shape[1]+2
        
        weights = torch.from_numpy(np.zeros((3,3,height, width))).float().to(self.device)
        active_point_img_coords = [np.random.choice(range(2,self.img_shape[0]-2),size=None), np.random.choice(range(2,self.img_shape[1]-2),size=None)]
        
        case = cases[case_id]
        if case=="blocked":
            offsetx, offsety = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
            target_offsetx, target_offsety = offsetx, offsety
            weights_active_point_img_coords = [active_point_img_coords[0]+1, active_point_img_coords[1]+1]
            weights[offsetx, offsety, weights_active_point_img_coords[0],weights_active_point_img_coords[1]] = 1.
            target_pointx = active_point_img_coords[0]+target_offsetx-1
            target_pointy = active_point_img_coords[1]+target_offsety-1

            target = torch.from_numpy(np.zeros((self.img_shape[0], self.img_shape[1]))).float().to(self.device)
            target[target_pointx, target_pointy] = 1.
            
            allow = self.smap3x3.compute_allow_matrix(weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            actual = self.smap3x3.prepare_flows_for_coord(allow, weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            offset = np.random.choice(range(3*3),size=None)
            referenced_point_x = active_point_img_coords[0]-(offset//3)
            referenced_point_y = active_point_img_coords[1]-(offset%3)
            try:
                if offset==((offsetx*3)+offsety):
                    assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == 0, "Flows preparing test failed. "
                else:
                    if (referenced_point_x>0) and (referenced_point_x<(self.img_shape[0]-1)) and (referenced_point_y>0) and (referenced_point_y<(self.img_shape[1]-1)):
                        assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == 1, "Flows preparing test failed. "
                    else:
                        assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == 0, "Flows preparing test failed. "
            except Exception as e:
                print(f"offset: {offset}")
                print(f"active_point_img_coords: {active_point_img_coords}")
                print(f"offsetx, offsety: {(offsetx, offsety)}")
                print(f"target_offsetx, target_offsety: {(target_offsetx, target_offsety)}")
                print(f"actual: {actual}")
                raise e
        else:
            offsetx, offsety = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
            target_offsetx, target_offsety = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
            weights_active_point_img_coords = [active_point_img_coords[0]+1, active_point_img_coords[1]+1]
            weights[offsetx, offsety, weights_active_point_img_coords[0],weights_active_point_img_coords[1]] = 1.
            target_pointx = active_point_img_coords[0]+target_offsetx-1
            target_pointy = active_point_img_coords[1]+target_offsety-1

            target = torch.from_numpy(np.zeros((self.img_shape[0], self.img_shape[1]))).float().to(self.device)
            target[target_pointx, target_pointy] = 1.
            
            allow = self.smap3x3.compute_allow_matrix(weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            actual = self.smap3x3.prepare_flows_for_coord(allow, weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            offset = np.random.choice(range(3*3),size=None)
            referenced_point_x = active_point_img_coords[0]-(offset//3)
            referenced_point_y = active_point_img_coords[1]-(offset%3)
            try:
                if offsetx!=target_offsetx or offsety!=target_offsety:
                    if (referenced_point_x>0) and (referenced_point_x<(self.img_shape[0]-1)) and (referenced_point_y>0) and (referenced_point_y<(self.img_shape[1]-1)):
                        assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == 1, "Flows preparing test failed. "
                    else:
                        assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == 0, "Flows preparing test failed. "
                else:
                    if offset==((offsetx*3)+offsety):
                        assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == 0, "Flows preparing test failed. "
                    else:
                        assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == 1, "Flows preparing test failed. "
            except Exception as e:
                print(f"offset: {offset}")
                print(f"active_point_img_coords: {active_point_img_coords}")
                print(f"offsetx, offsety: {(offsetx, offsety)}")
                print(f"target_offsetx, target_offsety: {(target_offsetx, target_offsety)}")
                print(f"actual: {actual}")
                raise e
                
    def test_prepare_flows_for_mask(self):
        cases = ["blocked", "random"]
        case_id = np.random.choice(len(cases),size=None)
        
        # With the assumption that the mask at the activated point is 1. 
        height = self.img_shape[0]+2
        width = self.img_shape[1]+2
        
        weights = torch.from_numpy(np.zeros((3,3,height, width))).float().to(self.device)
        active_point_img_coords = [np.random.choice(range(2,self.img_shape[0]-2),size=None), np.random.choice(range(2,self.img_shape[1]-2),size=None)]
        is_activated = True
            
        case = cases[case_id]
        if case=="blocked":
            offsetx, offsety = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
            target_offsetx, target_offsety = offsetx, offsety
            weights_active_point_img_coords = [active_point_img_coords[0]+1, active_point_img_coords[1]+1]
            if is_activated:
                weights[offsetx, offsety, weights_active_point_img_coords[0],weights_active_point_img_coords[1]] = 1.
            target_pointx = active_point_img_coords[0]+target_offsetx-1
            target_pointy = active_point_img_coords[1]+target_offsety-1

            target = torch.from_numpy(np.zeros((self.img_shape[0], self.img_shape[1]))).float().to(self.device)
            target[target_pointx, target_pointy] = 1.
            
            allow = self.smap3x3.compute_allow_matrix(weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            actual = self.smap3x3.prepare_flows_for_mask(allow, weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            offset = np.random.choice(range(3*3),size=None)
            try:
                assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == (((self.img_shape[0]-2)*(self.img_shape[1]-2)-1) if is_activated else ((self.img_shape[0]-2)*(self.img_shape[1]-2))), "Flows preparing test failed. "
            except Exception as e:
                print(f"is_activated: {is_activated}")
                print(f"offset: {offset}")
                print(f"active_point_img_coords: {active_point_img_coords}")
                print(f"offsetx, offsety: {(offsetx, offsety)}")
                print(f"target_offsetx, target_offsety: {(target_offsetx, target_offsety)}")
                print(f"actual: {actual}")
                raise e
        else:
            offsetx, offsety = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
            target_offsetx, target_offsety = np.random.choice(range(3),size=None), np.random.choice(range(3),size=None)
            weights_active_point_img_coords = [active_point_img_coords[0]+1, active_point_img_coords[1]+1]
            if is_activated:
                weights[offsetx, offsety, weights_active_point_img_coords[0],weights_active_point_img_coords[1]] = 1.
            target_pointx = active_point_img_coords[0]+target_offsetx-1
            target_pointy = active_point_img_coords[1]+target_offsety-1

            target = torch.from_numpy(np.zeros((self.img_shape[0], self.img_shape[1]))).float().to(self.device)
            target[target_pointx, target_pointy] = 1.
            
            allow = self.smap3x3.compute_allow_matrix(weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            actual = self.smap3x3.prepare_flows_for_mask(allow, weights.reshape(1,1,3,3,1, height, width), target.reshape(1,1,1,1, self.img_shape[0], self.img_shape[1]))
            
            offset = np.random.choice(range(3*3),size=None)
            try:
                if offsetx!=target_offsetx or offsety!=target_offsety:
                    assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == ((self.img_shape[0]-2)*(self.img_shape[1]-2)), "Flows preparing test failed. "
                else:
                    assert torch.sum(actual[0,offset,:,:]).reshape(-1).cpu().numpy()[0] == (((self.img_shape[0]-2)*(self.img_shape[1]-2)-1) if is_activated else ((self.img_shape[0]-2)*(self.img_shape[1]-2))), "Flows preparing test failed. "
            except Exception as e:
                print(f"is_activated: {is_activated}")
                print(f"offset: {offset}")
                print(f"active_point_img_coords: {active_point_img_coords}")
                print(f"offsetx, offsety: {(offsetx, offsety)}")
                print(f"target_offsetx, target_offsety: {(target_offsetx, target_offsety)}")
                print(f"actual: {actual}")
                raise e
    

if __name__ == "__main__":
    unittest.main()



import unittest
import os
import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
import pickle
from smap import *
from tools.testing.vtest.vtest_types import *


class SMap3x3VTestCase(unittest.TestCase):
    def setUp(self):
        self.device = "cpu"
        self.input_mask = torch.from_numpy(np.load("./tests/vtest_data/smap3x3/input.npy"))
        self.camera = np.array([[2304.5479, 0,  1686.2379], 
                                [0, 2305.8757, -0.0151],
                                [0, 0, 1.]], dtype=np.float32)
        self.panel = list(np.where(np.ones([self.input_mask.shape[0], self.input_mask.shape[1]])))
        self.panel[0] = self.panel[0] + .5
        self.panel[1] = self.panel[1] + .5
        self.smap3x3 = SMap3x3(self.input_mask.shape[0], self.input_mask.shape[1], self.camera, self.device).to(self.device)
        try:
            if not os.path.exists("./tests/vtest_data/output"):
                os.mkdir("./tests/vtest_data/output")
        except OSError as error:
            print(error)
        
    def test_in_x_1st_stage(self):
        test_types = ["above_left", "above", "above_right", "right", "below_right", "below", "below_left", "left"]
        test_id = np.random.choice(range(len(test_types)), size=None)
        test_type = test_types[test_id]
        smap3x3 = self.smap3x3.to(self.device)
        self.vtestcase = TestCase(name=f"test_{test_type}_target", testbot_in=TestBot_In(), testbot_out=TestBot_Out(), testbot_target=TestBot_Target())
        
        z = 1e3*np.random.rand(1)[0]
        input_repr = (z*(self.input_mask))
        input_repr = utils.to_3d(input_repr.reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]), self.input_mask.shape[0], self.input_mask.shape[1], self.panel, self.input_mask.shape, self.input_mask.shape, self.smap3x3.camera_matrix_inv, self.device)
        input_repr_x = nn.Parameter(input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,:1,:,:], requires_grad=True).to(self.device)
        input_repr_y = nn.Parameter(input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,1:2,:,:], requires_grad=True).to(self.device)
        input_repr_z = (input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,2:3,:,:]).to(self.device)
        
        input_mask = self.input_mask.reshape(1,1,1, self.input_mask.shape[0], self.input_mask.shape[1])
        input_mask = nn.Parameter(input_mask, requires_grad=True).to(self.device)
        
        target_repr = torch.from_numpy(np.load(f"./tests/vtest_data/smap3x3/{test_type}_target.npy"))
        target_repr = (1.*(target_repr)).reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]).to(self.device)
        target_repr = self.vtestcase.testbot_target(target_repr).detach()
        
        
        smap3x3.zero_grad()
        input_repr_x = self.vtestcase.testbot_in(input_repr_x, input_mask)
        weights = smap3x3(input_repr_x, input_repr_y, input_repr_z, input_mask, target_repr, self.input_mask.shape).reshape(1,-1, self.input_mask.shape[0], self.input_mask.shape[1])
        weights = torch.abs(self.vtestcase.testbot_out(weights)+1e-7)
        loss_m = torch.abs(weights-target_repr)
    
        loss_m = loss_m.reshape(1, -1).sum(dim=1)

        loss = torch.mean(loss_m)
        loss.backward()
        
        with open(f'./tests/vtest_data/output/test_{test_type}_target_flow_info.pkl', 'rb') as file:
            flow_data = pickle.load(file)
            actual = ((flow_data["activation_gradients"])["in"] != 0).astype(int)
            
        expected = np.load(f"./tests/vtest_data/smap3x3/input.npy").astype(int)
        if test_id==1 or test_id==5:
            expected = 0.*expected
        
        try:
            np.testing.assert_array_equal(actual, expected,
                                       err_msg=f"Gradient test with {test_type} target failed.")
        except Exception as e:
            print(f"z: {z}")
            print(f"actual: {actual}")
            print(f"expected: {expected}")
            raise e
            
    def test_in_x_2st_stage(self):
        test_type_param_dict = {"above_left": [0,0], "above": [0,1], "above_right": [0,2], "right": [1,0], "below_right": [2,0], "below": [2,1], "below_left": [2,2], "left": [1,2], "still": [1,1]}
        test_types = ["above_left", "above", "above_right", "right", "below_right", "below", "below_left", "left", "still"]
        test_id = [np.random.choice(range(len(test_types)), size=None), np.random.choice(range(len(test_types)), size=None)]
        test_type = [test_types[test_id[0]], test_types[test_id[1]]]
        smap3x3 = self.smap3x3.to(self.device)
        self.vtestcase = TestCase(name=f"test_{test_type[0]}_{test_type[1]}_target", testbot_in=TestBot_In(), testbot_out=TestBot_Out(), testbot_target=TestBot_Target())
        
        z = 1e3*np.random.rand(1)[0]
        offsetx, offsety = test_type_param_dict[test_type[0]]
        input_repr = (z*(self.input_mask))
        input_repr = utils.to_3d(input_repr.reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]), self.input_mask.shape[0], self.input_mask.shape[1], self.panel, self.input_mask.shape, self.input_mask.shape, self.smap3x3.camera_matrix_inv, self.device)
        input_repr = input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])
        active_point_img_coords = np.where(self.input_mask!=0)
        temp = (input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]]).cpu().numpy()
        input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]] = 0.*(input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]])
        input_repr[0,0,:,active_point_img_coords[0]+offsetx-1, active_point_img_coords[1]+offsety-1] = 0.+torch.from_numpy(temp)
        input_repr_x = nn.Parameter(input_repr[:,:,:1,:,:], requires_grad=True).to(self.device)
        input_repr_y = nn.Parameter(input_repr[:,:,1:2,:,:], requires_grad=True).to(self.device)
        input_repr_z = (input_repr[:,:,2:3,:,:]).to(self.device)
        
        input_mask = self.input_mask.reshape(1,1,1, self.input_mask.shape[0], self.input_mask.shape[1])
        temp = (input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]]).cpu().numpy()
        input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]] = 0*(input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]])
        input_mask[0,0,:,active_point_img_coords[0]+offsetx-1, active_point_img_coords[1]+offsety-1] = 0.+torch.from_numpy(temp)
        input_mask = nn.Parameter(input_mask, requires_grad=True).to(self.device)
        
        target_repr = torch.from_numpy(np.load(f"./tests/vtest_data/smap3x3/{test_type[1]}_target.npy"))
        target_repr = (1.*(target_repr)).reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]).to(self.device)
        target_repr = self.vtestcase.testbot_target(target_repr).detach()
        
        smap3x3.zero_grad()
        input_repr_x = self.vtestcase.testbot_in(input_repr_x, input_mask)
        weights = smap3x3(input_repr_x, input_repr_y, input_repr_z, input_mask, None, self.input_mask.shape)
        weights = smap3x3.go(weights[:,None,:,:,:], target_repr, self.input_mask.shape).reshape(1,-1, self.input_mask.shape[0], self.input_mask.shape[1])
        weights = torch.abs(self.vtestcase.testbot_out(weights)+1e-7)
        loss_m = torch.abs(weights-target_repr)
    
        loss_m = loss_m.reshape(1, -1).sum(dim=1)

        loss = torch.mean(loss_m)
        loss.backward()
        
        with open(f'./tests/vtest_data/output/test_{test_type[0]}_{test_type[1]}_target_flow_info.pkl', 'rb') as file:
            flow_data = pickle.load(file)
            actual = ((flow_data["activation_gradients"])["in"] != 0).astype(int)
            
        expected = input_mask.detach().reshape(input_mask.shape[-2], input_mask.shape[-1]).cpu().numpy()
        if test_id[1]==1 or test_id[1]==5 or test_types[test_id[1]]=="still":
            expected = 0.*expected
        
        try:
            np.testing.assert_array_equal(actual, expected,
                                       err_msg=f"Gradient test with {test_type} target failed.")
        except Exception as e:
            print(f"offsetx, offsety: {(offsetx, offsety)}")
            print(f"z: {z}")
            print(f"actual: {actual}")
            print(f"expected: {expected}")
            raise e
            
    def test_in_y_1st_stage(self):
        test_types = ["above_left", "above", "above_right", "right", "below_right", "below", "below_left", "left"]
        test_id = np.random.choice(range(len(test_types)), size=None)
        test_type = test_types[test_id]
        smap3x3 = self.smap3x3.to(self.device)
        self.vtestcase = TestCase(name=f"test_{test_type}_target", testbot_in=TestBot_In(), testbot_out=TestBot_Out(), testbot_target=TestBot_Target())
        
        z = 1e3*np.random.rand(1)[0]
        input_repr = (z*(self.input_mask))
        input_repr = utils.to_3d(input_repr.reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]), self.input_mask.shape[0], self.input_mask.shape[1], self.panel, self.input_mask.shape, self.input_mask.shape, self.smap3x3.camera_matrix_inv, self.device)
        input_repr_x = nn.Parameter(input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,:1,:,:], requires_grad=True).to(self.device)
        input_repr_y = nn.Parameter(input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,1:2,:,:], requires_grad=True).to(self.device)
        input_repr_z = (input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,2:3,:,:]).to(self.device)
        
        input_mask = self.input_mask.reshape(1,1,1, self.input_mask.shape[0], self.input_mask.shape[1])
        input_mask = nn.Parameter(input_mask, requires_grad=True).to(self.device)
        
        target_repr = torch.from_numpy(np.load(f"./tests/vtest_data/smap3x3/{test_type}_target.npy"))
        target_repr = (1.*(target_repr)).reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]).to(self.device)
        target_repr = self.vtestcase.testbot_target(target_repr).detach()
        
        
        smap3x3.zero_grad()
        input_repr_y = self.vtestcase.testbot_in(input_repr_y, input_mask)
        weights = smap3x3(input_repr_x, input_repr_y, input_repr_z, input_mask, target_repr, self.input_mask.shape).reshape(1,-1, self.input_mask.shape[0], self.input_mask.shape[1])
        weights = torch.abs(self.vtestcase.testbot_out(weights)+1e-7)
        loss_m = torch.abs(weights-target_repr)
    
        loss_m = loss_m.reshape(1, -1).sum(dim=1)

        loss = torch.mean(loss_m)
        loss.backward()
        
        with open(f'./tests/vtest_data/output/test_{test_type}_target_flow_info.pkl', 'rb') as file:
            flow_data = pickle.load(file)
            actual = ((flow_data["activation_gradients"])["in"] != 0).astype(int)
            
        expected = np.load(f"./tests/vtest_data/smap3x3/input.npy").astype(int)
        if test_id==3 or test_id==7:
            expected = 0.*expected
        
        try:
            np.testing.assert_array_equal(actual, expected,
                                       err_msg=f"Gradient test with {test_type} target failed.")
        except Exception as e:
            print(f"z: {z}")
            print(f"actual: {actual}")
            print(f"expected: {expected}")
            raise e
            
    def test_in_y_2st_stage(self):
        test_type_param_dict = {"above_left": [0,0], "above": [0,1], "above_right": [0,2], "right": [1,0], "below_right": [2,0], "below": [2,1], "below_left": [2,2], "left": [1,2], "still": [1,1]}
        test_types = ["above_left", "above", "above_right", "right", "below_right", "below", "below_left", "left", "still"]
        test_id = [np.random.choice(range(len(test_types)), size=None), np.random.choice(range(len(test_types)), size=None)]
        test_type = [test_types[test_id[0]], test_types[test_id[1]]]
        smap3x3 = self.smap3x3.to(self.device)
        self.vtestcase = TestCase(name=f"test_{test_type[0]}_{test_type[1]}_target", testbot_in=TestBot_In(), testbot_out=TestBot_Out(), testbot_target=TestBot_Target())
        
        z = 1e3*np.random.rand(1)[0]
        offsetx, offsety = test_type_param_dict[test_type[0]]
        input_repr = (z*(self.input_mask))
        input_repr = utils.to_3d(input_repr.reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]), self.input_mask.shape[0], self.input_mask.shape[1], self.panel, self.input_mask.shape, self.input_mask.shape, self.smap3x3.camera_matrix_inv, self.device)
        input_repr = input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])
        active_point_img_coords = np.where(self.input_mask!=0)
        temp = (input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]]).cpu().numpy()
        input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]] = 0.*(input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]])
        input_repr[0,0,:,active_point_img_coords[0]+offsetx-1, active_point_img_coords[1]+offsety-1] = 0.+torch.from_numpy(temp)
        input_repr_x = nn.Parameter(input_repr[:,:,:1,:,:], requires_grad=True).to(self.device)
        input_repr_y = nn.Parameter(input_repr[:,:,1:2,:,:], requires_grad=True).to(self.device)
        input_repr_z = (input_repr[:,:,2:3,:,:]).to(self.device)
        
        input_mask = self.input_mask.reshape(1,1,1, self.input_mask.shape[0], self.input_mask.shape[1])
        temp = (input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]]).cpu().numpy()
        input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]] = 0*(input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]])
        input_mask[0,0,:,active_point_img_coords[0]+offsetx-1, active_point_img_coords[1]+offsety-1] = 0.+torch.from_numpy(temp)
        input_mask = nn.Parameter(input_mask, requires_grad=True).to(self.device)
        
        target_repr = torch.from_numpy(np.load(f"./tests/vtest_data/smap3x3/{test_type[1]}_target.npy"))
        target_repr = (1.*(target_repr)).reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]).to(self.device)
        target_repr = self.vtestcase.testbot_target(target_repr).detach()
        
        smap3x3.zero_grad()
        input_repr_y = self.vtestcase.testbot_in(input_repr_y, input_mask)
        weights = smap3x3(input_repr_x, input_repr_y, input_repr_z, input_mask, None, self.input_mask.shape)
        weights = smap3x3.go(weights[:,None,:,:,:], target_repr, self.input_mask.shape).reshape(1,-1, self.input_mask.shape[0], self.input_mask.shape[1])
        weights = torch.abs(self.vtestcase.testbot_out(weights)+1e-7)
        loss_m = torch.abs(weights-target_repr)
    
        loss_m = loss_m.reshape(1, -1).sum(dim=1)

        loss = torch.mean(loss_m)
        loss.backward()
        
        with open(f'./tests/vtest_data/output/test_{test_type[0]}_{test_type[1]}_target_flow_info.pkl', 'rb') as file:
            flow_data = pickle.load(file)
            actual = ((flow_data["activation_gradients"])["in"] != 0).astype(int)
            
        expected = input_mask.detach().reshape(input_mask.shape[-2], input_mask.shape[-1]).cpu().numpy()
        if test_id[1]==3 or test_id[1]==7 or test_types[test_id[1]]=="still":
            expected = 0.*expected
        
        try:
            np.testing.assert_array_equal(actual, expected,
                                       err_msg=f"Gradient test with {test_type} target failed.")
        except Exception as e:
            print(f"offsetx, offsety: {(offsetx, offsety)}")
            print(f"z: {z}")
            print(f"actual: {actual}")
            print(f"expected: {expected}")
            raise e
    
    def test_in_r_1st_stage(self):
        testcase_name = "test_in_r_1st_stage"
        test_types = ["above_left", "above", "above_right", "right", "below_right", "below", "below_left", "left", "still"]
        test_id = np.random.choice(range(len(test_types)), size=None)
        test_type = test_types[test_id]
        
        smap3x3 = self.smap3x3.to(self.device)
        self.vtestcase = TestCase(name=f"{testcase_name}_{test_type}", testbot_in=TestBot_In(), testbot_out=TestBot_Out(), testbot_target=TestBot_Target())
        
        z = 1e3*np.random.rand(1)[0]
        input_repr = (z*(self.input_mask))
        input_repr = utils.to_3d(input_repr.reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]), self.input_mask.shape[0], self.input_mask.shape[1], self.panel, self.input_mask.shape, self.input_mask.shape, self.smap3x3.camera_matrix_inv, self.device)
        input_repr_x = nn.Parameter(input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,:1,:,:], requires_grad=True).to(self.device)
        input_repr_y = nn.Parameter(input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,1:2,:,:], requires_grad=True).to(self.device)
        input_repr_z = (input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])[:,:,2:3,:,:]).to(self.device)
        
        input_mask = self.input_mask.reshape(1,1,1, self.input_mask.shape[0], self.input_mask.shape[1])
        input_mask = nn.Parameter(input_mask, requires_grad=True).to(self.device)
        
        target_repr = torch.from_numpy(np.load(f"./tests/vtest_data/smap3x3/{test_type}_target.npy"))
        if test_type=="still":
            target_repr = torch.from_numpy(np.load(f"./tests/vtest_data/smap3x3/input.npy"))
        target_repr = (1.*(target_repr)).reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]).to(self.device)
        target_repr = self.vtestcase.testbot_target(target_repr)
        
        
        smap3x3.zero_grad()
        input_mask = self.vtestcase.testbot_in(input_mask, input_mask)
        weights = smap3x3(input_repr_x, input_repr_y, input_repr_z, input_mask, target_repr, self.input_mask.shape).reshape(1,-1, self.input_mask.shape[0], self.input_mask.shape[1])
        weights = torch.abs(self.vtestcase.testbot_out(weights)+1e-7)
        loss_m = torch.abs(weights-target_repr)
    
        loss_m = loss_m.reshape(1, -1).sum(dim=1)

        loss = torch.mean(loss_m)
        loss.backward()
        
        is_negative_check = np.random.choice([True, False], size=None)
        with open(f'./tests/vtest_data/output/{testcase_name}_{test_type}_flow_info.pkl', 'rb') as file:
            flow_data = pickle.load(file)
            actual = ((flow_data["activation_gradients"])["in"])
        if is_negative_check:
            actual = (actual < 0).astype(int)
            expected = np.load(f"./tests/vtest_data/smap3x3/{testcase_name}_{test_type}_target_expected.npy").astype(int)
        else:
            actual = (actual > 0).astype(int)
            expected = np.load("./tests/vtest_data/smap3x3/input.npy").astype(int)
            if test_type=="still":
                expected = 0.*expected
        
        try:
            np.testing.assert_array_equal(actual, expected,
                                       err_msg=f"Gradient test with {test_type} target failed.")
        except Exception as e:
            print(f"z: {z}")
            print(f"actual: {actual}")
            print(f"expected: {expected}")
            raise e
            
    def test_in_r_2st_stage(self):
        testcase_name = "test_in_r_2st_stage"
        test_type_param_dict = {"above_left": [0,0], "above": [0,1], "above_right": [0,2], "right": [1,0], "below_right": [2,0], "below": [2,1], "below_left": [2,2], "left": [1,2], "still": [1,1]}
        test_types = ["above_left", "above", "above_right", "right", "below_right", "below", "below_left", "left", "still"]
        test_id = [np.random.choice(range(len(test_types)), size=None), np.random.choice(range(len(test_types)), size=None)]
        test_type = [test_types[test_id[0]], test_types[test_id[1]]]
        smap3x3 = self.smap3x3.to(self.device)
        self.vtestcase = TestCase(name=f"{testcase_name}_{test_type[0]}_{test_type[1]}_target", testbot_in=TestBot_In(), testbot_out=TestBot_Out(), testbot_target=TestBot_Target())
        
        z = 1e3*np.random.rand(1)[0]
        offsetx, offsety = test_type_param_dict[test_type[0]]
        input_repr = (z*(self.input_mask))
        input_repr = utils.to_3d(input_repr.reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]), self.input_mask.shape[0], self.input_mask.shape[1], self.panel, self.input_mask.shape, self.input_mask.shape, self.smap3x3.camera_matrix_inv, self.device)
        input_repr = input_repr.reshape(1,1,3, self.input_mask.shape[0], self.input_mask.shape[1])
        active_point_img_coords = np.where(self.input_mask!=0)
        temp = (input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]]).cpu().numpy()
        input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]] = 0.*(input_repr[0,0,:,active_point_img_coords[0], active_point_img_coords[1]])
        input_repr[0,0,:,active_point_img_coords[0]+offsetx-1, active_point_img_coords[1]+offsety-1] = 0.+torch.from_numpy(temp)
        input_repr_x = nn.Parameter(input_repr[:,:,:1,:,:], requires_grad=True).to(self.device)
        input_repr_y = nn.Parameter(input_repr[:,:,1:2,:,:], requires_grad=True).to(self.device)
        input_repr_z = (input_repr[:,:,2:3,:,:]).to(self.device)
        
        input_mask = self.input_mask.reshape(1,1,1, self.input_mask.shape[0], self.input_mask.shape[1])
        temp = (input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]]).cpu().numpy()
        input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]] = 0*(input_mask[0,0,:,active_point_img_coords[0], active_point_img_coords[1]])
        input_mask[0,0,:,active_point_img_coords[0]+offsetx-1, active_point_img_coords[1]+offsety-1] = 0.+torch.from_numpy(temp)
        input_mask = nn.Parameter(input_mask, requires_grad=True).to(self.device)
        
        target_repr = torch.from_numpy(np.load(f"./tests/vtest_data/smap3x3/{test_type[1]}_target.npy"))
        target_repr = (1.*(target_repr)).reshape(1,1, self.input_mask.shape[0], self.input_mask.shape[1]).to(self.device)
        target_repr = self.vtestcase.testbot_target(target_repr).detach()
        
        smap3x3.zero_grad()
        input_mask = self.vtestcase.testbot_in(input_mask, input_mask)
        weights = smap3x3(input_repr_x, input_repr_y, input_repr_z, input_mask, None, self.input_mask.shape)
        weights = smap3x3.go(weights[:,None,:,:,:], target_repr, self.input_mask.shape).reshape(1,-1, self.input_mask.shape[0], self.input_mask.shape[1])
        weights = torch.abs(self.vtestcase.testbot_out(weights)+1e-7)
        loss_m = torch.abs(weights-target_repr)
    
        loss_m = loss_m.reshape(1, -1).sum(dim=1)

        loss = torch.mean(loss_m)
        loss.backward()
        
        is_negative_check = np.random.choice([True, False], size=None)
        with open(f'./tests/vtest_data/output/{testcase_name}_{test_type[0]}_{test_type[1]}_target_flow_info.pkl', 'rb') as file:
            flow_data = pickle.load(file)
            actual = ((flow_data["activation_gradients"])["in"])
            
        expected_factor = (0 if test_type[1]=="still" else 1)
        expected_pos = [active_point_img_coords[0]+(offsetx-1), active_point_img_coords[1]+(offsety-1)]
        if is_negative_check:
            actual = (actual < 0).astype(int)
            expected = np.load(f"./tests/vtest_data/smap3x3/test_in_r_1st_stage_{test_type[1]}_target_expected.npy").astype(int)
            expected[expected_pos[0], expected_pos[1]] = 0
        else:
            actual = (actual > 0).astype(int)
            expected = 0.*np.load("./tests/vtest_data/smap3x3/input.npy").astype(int)
            expected[expected_pos[0], expected_pos[1]] = expected_factor
        
        try:
            np.testing.assert_array_equal(actual, expected,
                                       err_msg=f"Gradient test with {test_type} target failed.")
        except Exception as e:
            print(f"offsetx, offsety: {(offsetx, offsety)}")
            print(f"z: {z}")
            print(f"actual: {actual}")
            print(f"expected: {expected}")
            raise e
            

if __name__ == "__main__":
    unittest.main()


from setuptools import setup, find_packages
import os, re

def read_version():
    with open("smap/__init__.py", "r", encoding="utf8") as f:
        content = f.read()
    version_match = re.search(r"^__version__\s*=\s*['\"]([^'\"]+)['\"]", content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Could not find __version__ in __init__.py")

# Đọc file requirements.txt và loại bỏ các dòng trống hoặc comment
with open("requirements.txt", "r", encoding="utf-8") as req_file:
    install_requires = [
        line.strip() for line in req_file
        if line.strip() and not line.startswith("#")
    ]
    
# Đọc nội dung README.md
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="smap-torch",  # Tên package
    version=read_version(),  # Phiên bản khởi đầu
    description="An open source pytorch library for spatial mapping based on 2D representations",  # Mô tả ngắn
    long_description=long_description,
    long_description_content_type="text/markdown",  # hoặc "text/x-rst"
    author="Thien An L. Nguyen",
    author_email="thienannguyen.cv@gmail.com",
    url="https://github.com/thienannguyen-cv/SMap",
    packages=find_packages(),  # Tự động tìm các package con
    install_requires=install_requires,
    license="Apache-2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6.7",
)



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from smap import specials

def flip(x, dim):
    dim = x.dim() + dim if dim < 0 else dim
    return x[tuple(slice(None, None) if i != dim
             else torch.arange(x.size(i)-1, -1, -1).long()
             for i in range(x.dim()))]
    
class SMap3x3(nn.Module):
    def __init__(self, window_h, window_w, camera_matrix, device):
        super(SMap3x3,self).__init__()
        self.vtest_service = None
        self.window_h = window_h
        self.window_w = window_w
        self.camera_matrix = nn.Parameter(torch.from_numpy(camera_matrix), requires_grad=False)
        self.camera_matrix_inv = nn.Parameter(torch.from_numpy(np.linalg.inv(camera_matrix)), requires_grad=False)
        self.device = device
        self.grad_grouped_key = None
        self.sm = nn.Softmax(dim=2)
    
    def agg(self, x, ind=None, factor=None):
        fct = 0.
        if factor is not None:
            fct = factor
        x = x + ((x>0.).float()-1.)*(-fct)
        
        sizes = list(x.size())
        sizes[2] = 3
        sizes[3] = 3
        
        x = x.reshape(*sizes)
        
        sizes[2] = 1
        sizes[3] = 1
        
        def abs_alignment(x, relx, rely, fct):
            x = 1.*(x[:,:,relx,rely,:,:,:])
            if relx<1:
                x = torch.cat([(x[:,:,:,1:,:]), torch.ones_like(x[:,:,:,:1,:])*fct],dim=-2)
            if relx>1:
                x = torch.cat([torch.ones_like(x[:,:,:,-1:,:])*fct, (x[:,:,:,:-1,:])],dim=-2)
            if rely<1:
                x = torch.cat([(x[:,:,:,:,1:]), torch.ones_like(x[:,:,:,:,:1])*fct],dim=-1)
            if rely>1:
                x = torch.cat([torch.ones_like(x[:,:,:,:,-1:])*fct, (x[:,:,:,:,:-1])],dim=-1)
            return x
        
        ys = []
        for i in range(3):
            for j in range(3):
                ys.append(abs_alignment(x, i, j, fct).reshape(*sizes))
        
        sizes[2] = 3*3
        sizes[3] = 1
        
        x = torch.cat(ys,dim=2).reshape(*sizes) # [y00,y01,y02,y10,y11,y12,y20,y21,y22]

        if ind is None:
            return x
        if sizes[4] == 4:
            sizes[4] = 1
            x0 = (x[:,:,:,:,:1,:,:]).reshape(*sizes)
            x1 = (x[:,:,:,:,1:2,:,:]).reshape(*sizes)
            x2 = (x[:,:,:,:,2:3,:,:]).reshape(*sizes)
            x3 = (x[:,:,:,:,3:,:,:]).reshape(*sizes)
            
            sizes[2] = 1
            sizes[3] = 1
            
            x0 = torch.sum(torch.where(ind,x0,torch.zeros_like(x0)),dim=2,keepdim=True)
            x1 = torch.sum(torch.where(ind,x1,torch.zeros_like(x1)),dim=2,keepdim=True)
            x2 = torch.sum(torch.where(ind,x2,torch.zeros_like(x2)),dim=2,keepdim=True)
            x3 = torch.sum(torch.where(ind,x3,torch.zeros_like(x3)),dim=2,keepdim=True)
            
            return torch.cat([x0.reshape(*sizes), x1.reshape(*sizes), x2.reshape(*sizes), x3.reshape(*sizes)], dim=4)
        
        sizes[4] = 1
        x0 = (x[:,:,:,:,:1,:,:]).reshape(*sizes)

        sizes[2] = 1
        sizes[3] = 1

        x0 = torch.sum(torch.where(ind,x0,torch.zeros_like(x0)),dim=2,keepdim=True)
        return x0.reshape(*sizes)
    
    def compute_allow_matrix(self, ws, tgt_repr):
        BATCH_SIZE, C_zoom ,height, width, w_zoom, h_zoom = ws.shape[0], ws.shape[1], ws.shape[-2], ws.shape[-1], tgt_repr.shape[-1], tgt_repr.shape[-2]
        
        ws = self.agg(ws).reshape(BATCH_SIZE,C_zoom,3*3,1,height, width)
        
        allow = torch.ones_like(ws)
        wsxtgt_repr = ((ws[:,:,:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)])==tgt_repr)
        # Dùng để đếm các điểm trên ws đã kích hoạt trùng khớp với tgt_repr. 
        wsxtgt_repr = wsxtgt_repr*(ws[:,:,:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)])
        # Nếu các điểm ứng với điểm trên tgt_rept đã được kích hoạt và số trùng khớp là khác 1. 
        wsxtgt_repr = (ws[:,:,:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)])*(torch.sum(wsxtgt_repr,dim=2,keepdim=True)==1.).float()
        
        allow[:,:,:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)] = wsxtgt_repr
        allow = self.agg(flip(allow,2).reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)).reshape(BATCH_SIZE,C_zoom,3*3,1,1,height, width)
        allow = torch.max(allow,dim=2,keepdim=True).values
        allow = 1.-allow
        return allow
    
    def prepare_flows_for_mask(self, allow, ws, tgt_repr):
        BATCH_SIZE, C_zoom ,height, width, w_zoom, h_zoom = ws.shape[0], ws.shape[1], ws.shape[-2], ws.shape[-1], tgt_repr.shape[-1], tgt_repr.shape[-2]
        
        allow = torch.cat([allow, allow, allow],dim=2)
        allow = torch.cat([allow, allow, allow],dim=3).reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)
        allow = (self.agg(allow).reshape(BATCH_SIZE,-1,height, width)[:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)]).reshape(BATCH_SIZE,-1,h_zoom, w_zoom)
        return allow
    
    def prepare_flows_for_coord(self, allow, ws, tgt_repr):
        BATCH_SIZE, C_zoom ,height, width, w_zoom, h_zoom = ws.shape[0], ws.shape[1], ws.shape[-2], ws.shape[-1], tgt_repr.shape[-1], tgt_repr.shape[-2]
        
        allow = self.prepare_flows_for_mask(allow, ws, tgt_repr)
        return allow*tgt_repr.reshape(BATCH_SIZE,1,h_zoom, w_zoom)
    
    def go(self, x, target_2Dr, original_size):
        x_, y, z, mask = x[:,:,:1,:,:], x[:,:,1:2,:,:], x[:,:,2:3,:,:], x[:,:,3:4,:,:]
        return self(x_, y, z, mask, target_2Dr, original_size)
    
    def forward(self, x_value, y_value, z_value, r_mask, target_2Dr, original_size):
        from smap import utils
        
        shapes = x_value.size()
        BATCH_SIZE, height, width = shapes[0], shapes[-2], shapes[-1]
        
        # 1. Prepare configuration for to_3d unit
        height = height + 2**0 + 2**0
        width = width + 2**0 + 2**0
        
        panels = list(np.where(np.ones([height, width])))
        offset_codes = ((height-original_size[0]), (width-original_size[1]))
        panels[0] = panels[0] - (offset_codes[0]//2) + .5
        panels[1] = panels[1] - (offset_codes[1]//2) + .5
        #######################
        
        
        # 2. Prepare input tensors
        x_value = torch.cat([torch.zeros_like(x_value[:,:,:,:,:(2**0)]), x_value, torch.zeros_like(x_value[:,:,:,:,:(2**0)])], dim=-1)
        x_value = torch.cat([torch.zeros_like(x_value[:,:,:,:(2**0),:]), x_value, torch.zeros_like(x_value[:,:,:,:(2**0),:])], dim=-2)
        
        y_value = torch.cat([torch.zeros_like(y_value[:,:,:,:,:(2**0)]), y_value, torch.zeros_like(y_value[:,:,:,:,:(2**0)])], dim=-1)
        y_value = torch.cat([torch.zeros_like(y_value[:,:,:,:(2**0),:]), y_value, torch.zeros_like(y_value[:,:,:,:(2**0),:])], dim=-2)
        
        z_value = torch.cat([torch.zeros_like(z_value[:,:,:,:,:(2**0)]), z_value, torch.zeros_like(z_value[:,:,:,:,:(2**0)])], dim=-1)
        z_value = torch.cat([torch.zeros_like(z_value[:,:,:,:(2**0),:]), z_value, torch.zeros_like(z_value[:,:,:,:(2**0),:])], dim=-2)
        
        r_mask = torch.cat([torch.zeros_like(r_mask[:,:,:,:,:(2**0)]), r_mask, torch.zeros_like(r_mask[:,:,:,:,:(2**0)])], dim=-1)
        r_mask = torch.cat([torch.zeros_like(r_mask[:,:,:,:(2**0),:]), r_mask, torch.zeros_like(r_mask[:,:,:,:(2**0),:])], dim=-2)
        
        z_values = z_value.reshape(BATCH_SIZE,-1,1,height, width)
        r_mask = r_mask.reshape(BATCH_SIZE,-1,1,height, width)
        x_z_value = (torch.cat([x_value, y_value, z_value], dim=2)).reshape(BATCH_SIZE,-1,3,height, width)
        grouped_key_x = x_value.reshape(BATCH_SIZE,-1,1,1,height*width)
        grouped_key_y = y_value.reshape(BATCH_SIZE,-1,1,1,height*width)
        #######################
        
        
        # 3. Prepare spatial placeholders for recifying gradients
        updated_key_z = utils.to_3d3x3(z_values.reshape(BATCH_SIZE,-1,height, width), height, width, panels, original_size, (self.window_h, self.window_w), self.camera_matrix_inv, self.device).permute(0,1,3,2).reshape(BATCH_SIZE,-1,3,3*3,height*width)
        query_x = (updated_key_z[:,:,:1,:,:]).detach()
        query_y = (updated_key_z[:,:,1:2,:,:]).detach()
        
        key_query = torch.sum(torch.abs(grouped_key_x-query_x)+torch.abs(grouped_key_y-query_y),dim=2).reshape(BATCH_SIZE,-1,3*3,height, width)
        #######################
        
        
        # 4. Setting proper tensor, named `weights_b', for differentiable rendering
        new_r_mask = torch.zeros_like(key_query)
        ind = torch.max(-key_query,dim=2,keepdim=True).indices
        ind_mask = F.one_hot(ind, num_classes=3*3).reshape(BATCH_SIZE,-1,height, width,3*3).permute(0,1,4,2,3).reshape(BATCH_SIZE,-1,3*3,height, width)
        new_r_mask[ind_mask>.5] = 1.
        new_r_mask = torch.where((r_mask+(query_x[:,:,0,:,:]).reshape(BATCH_SIZE,-1,3*3,height, width)*0.)>specials.OFF_THRESH, new_r_mask, torch.zeros_like(new_r_mask))
        new_r_mask[:,:,4,:,:] = torch.where(r_mask.reshape(BATCH_SIZE,-1,height, width)>specials.OFF_THRESH, new_r_mask[:,:,4,:,:], torch.ones_like(new_r_mask[:,:,4,:,:]))
        weights_b = (new_r_mask*(r_mask>specials.OFF_THRESH).float()).reshape(BATCH_SIZE,-1,3,3,height, width)
        new_x_z_value = torch.einsum('bcsthw,bczhw->bcstzhw', weights_b.detach(), x_z_value)
        new_z_values = torch.einsum('bcsthw,bczhw->bcstzhw', weights_b, z_values).detach()
        new_r_mask = (new_r_mask*r_mask).reshape(BATCH_SIZE,-1,3,3,1,height, width)
        new_x_z_mask_value = torch.cat([new_x_z_value, new_r_mask], dim=4)
        
        new_x_z_value = None
        weights = None
        
        ind = self.agg(new_z_values, factor=1e7)
        ind = torch.min(ind,dim=2,keepdim=True).indices
        ind = torch.where(torch.sum(self.agg(new_r_mask, factor=0.),dim=2,keepdim=True)>.5, ind, 0*ind+4)
        ind = F.one_hot(ind, num_classes=3*3).reshape(BATCH_SIZE,-1,1,1,height, width,3*3).permute(0,1,6,2,3,4,5).reshape(BATCH_SIZE,-1,3*3,1,1,height, width)
        ind = (ind>.5)
        weights = self.agg(new_x_z_mask_value, ind=ind).reshape(-1,4,height, width)
        #######################
            
            
        # 7. Triggering gradient at the origins of the image rectification
        if target_2Dr is not None:
            shapes = target_2Dr.size()
            BATCH_SIZE, C_zoom, h_zoom, w_zoom = shapes[0], shapes[1], shapes[-2], shapes[-1]
            
            weights_b = weights_b.reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)
            new_r_mask = new_r_mask.reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)
            r_mask = r_mask.reshape(BATCH_SIZE,C_zoom,1,1,1,height, width)
            target_2Dr = torch.max(target_2Dr.reshape(BATCH_SIZE,-1,1,1,h_zoom, w_zoom),dim=1,keepdim=True).values
            
            allow = self.compute_allow_matrix(weights_b.detach(), target_2Dr).detach()
            
            a_weights = new_r_mask*(new_r_mask>specials.OFF_THRESH).float()*(1.-allow)
            a_weights = a_weights.reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)
            a_weights = self.agg(a_weights).reshape(BATCH_SIZE,-1,3*3,height, width) # 2nd diff
            n_weights = torch.zeros_like(new_r_mask)+.75*(r_mask.detach() - (2.*(r_mask>0.).float()-1.)*(r_mask-r_mask.detach()))*allow
            n_weights = n_weights.reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)
            n_weights = self.agg(n_weights).reshape(BATCH_SIZE,-1,3*3,height, width)
            n_weights = n_weights.detach()+torch.max(1.-(a_weights>specials.OFF_THRESH).float(),dim=2,keepdim=True).values*(n_weights-n_weights.detach())
            weights = (a_weights-n_weights).reshape(BATCH_SIZE,-1,height, width)
            new_r_mask = new_r_mask.reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)
            new_r_mask = self.agg(new_r_mask).reshape(BATCH_SIZE,-1,height, width) # 2nd diff
            
            key_query_grdf = -(key_query-key_query.detach()).reshape(BATCH_SIZE,C_zoom,3,3,1,height, width)
            key_query_grdf = self.agg(key_query_grdf).reshape(BATCH_SIZE,-1,height, width)
            
            mask_flow = self.prepare_flows_for_mask(allow, weights_b, target_2Dr)
            coord_flow = mask_flow*target_2Dr.reshape(BATCH_SIZE,1,h_zoom, w_zoom)
            
            
            weights = (weights[:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)]).reshape(BATCH_SIZE,-1,h_zoom, w_zoom)
            key_query_grdf = (key_query_grdf.reshape(BATCH_SIZE,-1,height, width)[:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)]).reshape(BATCH_SIZE,-1,h_zoom, w_zoom)
            new_r_mask = (new_r_mask[:,:,((height-h_zoom)//2):((height+h_zoom)//2),((width-w_zoom)//2):((width+w_zoom)//2)]).reshape(BATCH_SIZE,-1,h_zoom, w_zoom)
            
            target_2Dr = target_2Dr.reshape(BATCH_SIZE,1,h_zoom, w_zoom)
            weights_grdf = (weights-weights.detach())
            
            case = ((torch.abs(weights)>specials.OFF_THRESH).float()==target_2Dr).float()
            n_case = ((new_r_mask>specials.OFF_THRESH).float()==target_2Dr).float()
            weights = weights.detach() + (2.*(weights>0.).float()-1.)*((1.-case)*weights_grdf + (1.-n_case)*key_query_grdf*coord_flow.detach()) # apply attractive rectification for this implementation
        #######################
        
        return weights

class SMap(nn.Module):
    def __init__(self, n, window_h, window_w, camera_matrix, device):
        super(SMap,self).__init__()
        self.vtest_service = None
        self.n = n
        self.smap3x3 = SMap3x3(window_h, window_w, camera_matrix, device)
    
    def forward(self, x, target_2Dr=None, zoom=0):
        shapes = x.size()
        BATCH_SIZE, height, width = shapes[0], shapes[2], shapes[3]
        C_zoom = 2**(self.n+self.n)
        C_zoom_2 = 1
        height_zoom = height
        width_zoom = width
        for i in range(self.n):
            height_zoom = height_zoom // 2
            width_zoom = width_zoom // 2
            x = x.reshape(BATCH_SIZE,C_zoom_2,C_zoom_2,4,height_zoom,2, width_zoom,2).permute(0,5,1,7,2,3,4,6).reshape(BATCH_SIZE,-1,4,height_zoom, width_zoom)
            C_zoom_2 = C_zoom_2 * 2
            
        target = None
        if target_2Dr is not None:
            target_2Dr = target_2Dr.reshape(BATCH_SIZE,1,height_zoom,C_zoom_2, width_zoom,C_zoom_2).permute(0,3,5,1,2,4).reshape(BATCH_SIZE,C_zoom,1,height_zoom, width_zoom)
            if self.n==zoom:
                target = target_2Dr.reshape(BATCH_SIZE,-1,height_zoom, width_zoom)
        
        # x.shape
        # >>> torch.Size([16, 256, 4, 8, 16])
        x = x.reshape(x.size(0),-1,x.size(-3),x.size(-2),x.size(-1))
        x = self.smap3x3.go(x, target, (height_zoom, width_zoom))
        _, _, h_out, w_out = x.size()
        
        for i in range(self.n-zoom):
            C_zoom = C_zoom//4
            C_zoom_2 = C_zoom_2//2
            height_zoom = height_zoom*2
            width_zoom = width_zoom*2
            h_out = h_out*2
            w_out = w_out*2
            x = x.reshape(BATCH_SIZE,2,C_zoom_2,2,C_zoom_2,4,h_out//2, w_out//2).permute(0,2,4,5,6,1,7,3).reshape(BATCH_SIZE,C_zoom,4,h_out, w_out)
            target = None
            if target_2Dr is not None:
                target_2Dr = target_2Dr.reshape(BATCH_SIZE,2,C_zoom_2,2,C_zoom_2,1,height_zoom//2, width_zoom//2).permute(0,2,4,5,6,1,7,3).reshape(BATCH_SIZE,C_zoom,height_zoom, width_zoom)
                if i==(self.n-zoom-1):
                    target = target_2Dr.reshape(BATCH_SIZE,-1,height_zoom, width_zoom)
                    
            x = self.smap3x3.go(x, target, (height_zoom, width_zoom))
            _, _, h_out, w_out = x.size()
        
        return x


import numpy as np
import torch
import torch.nn.functional as F
from smap import specials


def to_3d3x3(z, height, width, panels, original_size, window_size, camera_matrix_inv, device):
    y_im, x_im = panels
    y_im, x_im = torch.from_numpy(y_im).reshape(height, width), torch.from_numpy(x_im).reshape(height, width)
    y_im = y_im * window_size[0] / original_size[0]
    x_im = x_im * window_size[1] / original_size[1]
    y_im, x_im = y_im.to(device), x_im.to(device)

    imp_co = torch.cat([torch.einsum('hw,bczhw->bczhw', x_im.float(), torch.ones_like(z.unsqueeze(2)).float()), torch.einsum('hw,bczhw->bczhw', y_im.float(), torch.ones_like(z.unsqueeze(2)).float()), torch.ones_like(z.unsqueeze(2))], 2)
    imp_co = F.unfold(imp_co.reshape(1, -1, height, width), kernel_size=(3,3), stride=(1,1), padding=(1,1), dilation=(1,1)).reshape(z.size(0),z.size(1),3,3*3,height,width)
    
    imp_co = torch.einsum('bchw,bczshw->bczshw', z.float(), imp_co.float()).reshape(z.size(0),z.size(1),3,3*3,-1)
    
    regr_co = torch.einsum('xz,yz->xy', imp_co.reshape(z.size(0),z.size(1),3,-1).permute(0,1,3,2).reshape(-1,3).float(), camera_matrix_inv.float())
    regr_co = regr_co.reshape(z.size(0),z.size(1),-1,3)
    return regr_co

def to_3d(z, height, width, panels, original_size, window_size, camera_matrix_inv, device):
    regr_co = to_3d3x3(z, height, width, panels, original_size, window_size, camera_matrix_inv, device).reshape(z.size(0),z.size(1),3,3,-1,3)
    return (regr_co[:,:,1,1,:,:]).permute(0,1,3,2).reshape(-1,3, height, width)

def recover_size(x, n, zoom):
    BATCH_SIZE, C_zoom, h_out, w_out = x.size()
    C_zoom = C_zoom//(3*3)
    C_zoom_2 = int(np.sqrt(C_zoom))
    x = (1.*x).reshape(BATCH_SIZE,-1,3*3,h_out, w_out)
    for i in range(zoom):
        C_zoom = C_zoom//4
        C_zoom_2 = C_zoom_2//2
        h_out = h_out*2
        w_out = w_out*2
        x = x.reshape(BATCH_SIZE,2,C_zoom_2,2,C_zoom_2,3*3,h_out//2, w_out//2).permute(0,2,4,5,6,1,7,3).reshape(BATCH_SIZE,C_zoom,3*3,h_out, w_out)
    return torch.max((x>specials.OFF_THRESH).float(),dim=2,keepdim=False).values.reshape(BATCH_SIZE,1,h_out, w_out)

def save_for_vtest(path,activation_gradients, gradient_flows, input_representation, target_representation):
    import pickle
    
    flow_info = {"activation_gradients": activation_gradients, 
                 "gradient_flows": gradient_flows}
    with open(f'{path}/flow_info.pkl', 'wb') as f:
                    pickle.dump(flow_info, f)
    np.save(f"{path}/input_representation.npy", input_representation)
    np.save(f"{path}/target_representation.npy", target_representation)
    